////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                           ManagedSegmentList.cc                            //
//                                                                            //
//             Container for managing segment lists used by hMon              //
//                                                                            //
//                            Shourov K. Chatterji                            //
//                                                                            //
//                                2008-Sep-04                                 //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

// $Id: ManagedSegmentList.cc 5251 2008-09-05 16:41:59Z shourov.chatterji $

#include "ManagedSegmentList.hh"
#include <algorithm>
#include <stdexcept>
#include <iostream>

/******************************************************************************/
/*                             ManagedSegmentList                             */
/******************************************************************************/

ManagedSegmentList::ManagedSegmentList()
{
}

/******************************************************************************/
/*                            ~ManagedSegmentList                             */
/******************************************************************************/

ManagedSegmentList::~ManagedSegmentList()
{
}

/******************************************************************************/
/*                                getFileName                                 */
/******************************************************************************/

const std::string&
ManagedSegmentList::getFileName(void) const
{
  return fileName;
}

/******************************************************************************/
/*                                getListName                              */
/******************************************************************************/

const std::string&
ManagedSegmentList::getListName(void) const
{
  return listName;
}

/******************************************************************************/
/*                             getListDescription                             */
/******************************************************************************/

const std::string&
ManagedSegmentList::getListDescription(void) const
{
  return listDescription;
}

/******************************************************************************/
/*                               isDispatchMode                               */
/******************************************************************************/

bool
ManagedSegmentList::isDispatchMode(std::string mode)
{
  std::transform(mode.begin(), mode.end(), mode.begin(),
              (int(*)(int)) std::tolower);
  if (dispatchModes.find(mode) != dispatchModes.end()) {
    return true;
  }
  else {
    return false;
  }
}

/******************************************************************************/
/*                             getMaximumDuration                             */
/******************************************************************************/

Interval
ManagedSegmentList::getMaximumDuration(void) const
{
  return(maximumDuration);
}

/******************************************************************************/
/*                             getRecentDuration                              */
/******************************************************************************/

Interval
ManagedSegmentList::getRecentDuration(void) const
{
  return(recentDuration);
}

/******************************************************************************/
/*                             getLastUpdateTime                              */
/******************************************************************************/

const Time&
ManagedSegmentList::getLastUpdateTime(void) const
{
  return(lastUpdateTime);
}

/******************************************************************************/
/*                              getWriteSegments                              */
/******************************************************************************/

const LockSegList&
ManagedSegmentList::getWriteSegments(void) const
{
  merge();
  return(writeSegments);
}

/******************************************************************************/
/*                             getUnsentSegments                              */
/******************************************************************************/

const LockSegList&
ManagedSegmentList::getUnsentSegments(void) const
{
  merge();
  return(unsentSegments);
}

/******************************************************************************/
/*                             getRecentSegments                              */
/******************************************************************************/

const LockSegList&
ManagedSegmentList::getRecentSegments(void) const
{
  return(recentSegments);
}

/******************************************************************************/
/*                                setFileName                                 */
/******************************************************************************/

void
ManagedSegmentList::setFileName(const std::string& name)
{
  fileName = name;
  return;
}

/******************************************************************************/
/*                                setListName                                 */
/******************************************************************************/

void
ManagedSegmentList::setListName(const std::string& name)
{
  listName = name;
  return;
}

/******************************************************************************/
/*                             setListDescription                             */
/******************************************************************************/

void
ManagedSegmentList::setListDescription(const std::string& description)
{
  listDescription = description;
  return;
}

/******************************************************************************/
/*                              addDispatchMode                               */
/******************************************************************************/

void
ManagedSegmentList::addDispatchMode(std::string mode)
{
  std::transform(mode.begin(), mode.end(), mode.begin(),
                 (int(*)(int)) std::tolower);
  if ((mode != "txt") &
      (mode != "xml") &
      (mode != "sql") &
      (mode != "none")) {
    std::cerr << "ERROR: invalid dispatch mode in "
              << "ManagedSegmentList::addDispatchMode()" << std::endl;
    throw std::runtime_error("Invalid dispatch mode");
  }
  if (mode == "none") {
    clearDispatchModes();
  }
  else if (!isDispatchMode(mode)) {
    dispatchModes.insert(mode);
  }
  return;
}

/******************************************************************************/
/*                             removeDispatchMode                             */
/******************************************************************************/

void
ManagedSegmentList::removeDispatchMode(std::string mode)
{
  std::transform(mode.begin(), mode.end(), mode.begin(),
                 (int(*)(int)) std::tolower);
  if (!isDispatchMode(mode)) {
    std::cerr << "WARNING: non-existent dispatch mode in "
              << "ManagedSegmentList::removeDispatchMode()" << std::endl;
    throw std::runtime_error("Non-existent dispatch mode");
  }
  else {
    dispatchModes.erase(mode);
  }
  return;
}

/******************************************************************************/
/*                             clearDispatchModes                             */
/******************************************************************************/

void
ManagedSegmentList::clearDispatchModes(void)
{
  dispatchModes.clear();
}

/******************************************************************************/
/*                             setMaximumDuration                             */
/******************************************************************************/

void
ManagedSegmentList::setMaximumDuration(const Interval& duration)
{
  maximumDuration = duration;
  return;
}

/******************************************************************************/
/*                             setRecentDuration                              */
/******************************************************************************/

void
ManagedSegmentList::setRecentDuration(const Interval& duration)
{
  recentDuration = duration;
  return;
}

/******************************************************************************/
/*                                   insert                                   */
/******************************************************************************/

void
ManagedSegmentList::insert(const LockSegment& segment)
{
  recentSegments.merge(segment);
  isMerged = false;
}

/******************************************************************************/
/*                                   insert                                   */
/******************************************************************************/

void
ManagedSegmentList::insert(const Time& startTime, const Time& stopTime)
{
  LockSegment segment(0, startTime, stopTime);
  insert(segment);
}

/******************************************************************************/
/*                                   insert                                   */
/******************************************************************************/

void
ManagedSegmentList::insert(const Time& startTime, const Interval& duration)
{
  insert(startTime, startTime + duration);
}

/******************************************************************************/
/*                                   insert                                   */
/******************************************************************************/

void
ManagedSegmentList::insert(const LockSegList& segmentList)
{
  recentSegments |= segmentList;
  isMerged = false;
}

/******************************************************************************/
/*                                   update                                   */
/******************************************************************************/

void
ManagedSegmentList::update(const Time& updateTime)
{
  if (updateTime < lastUpdateTime) {
    std::cerr << "ERROR: invalid update time in "
              << "ManagedSegmentList::update()" << std::endl;
    throw std::runtime_error("Invalid update time");
  }
  merge();
  mask(writeSegments, updateTime - maximumDuration);
  mask(recentSegments, updateTime - recentDuration);
  lastUpdateTime = updateTime;
  return;
}

/******************************************************************************/
/*                                  markSent                                  */
/******************************************************************************/

void
ManagedSegmentList::markSent(void)
{
  unsentSegments.clear();
  return;
}

/******************************************************************************/
/*                                   reset                                    */
/******************************************************************************/

void
ManagedSegmentList::reset()
{
  writeSegments.clear();
  unsentSegments.clear();
  recentSegments.clear();
  lastUpdateTime = Time(0);
}

/******************************************************************************/
/*                                    mask                                    */
/******************************************************************************/

void
ManagedSegmentList::mask(LockSegList& segmentList,
                         const Time& startTime,
                         const Time& stopTime) const
{
  LockSegment maskSegment(0, startTime, stopTime);
  LockSegList maskSegments;
  maskSegments.insert(maskSegment);
  segmentList &= maskSegments;
}

/******************************************************************************/
/*                                   merge                                    */
/******************************************************************************/

void
ManagedSegmentList::merge(void) const
{
  if (isMerged == false) {
    writeSegments |= recentSegments;
    unsentSegments |= recentSegments;
    isMerged = true;
  }
}
